/**
 * \file dissolve.c
 * \brief Dissolve Effect Demo - inverts pixels in a pseudo-random order
 *
 * lionel_debroux@yahoo.fr (TICT) 2010-2011
 *
 * Port to the Nspire (Ndless pre-2.0) of the TI-68k version made in 2000-2004
 * by Thomas Nussbaumer & Lionel Debroux.
 * Based on code from Mike Morton (Honolulu, Hawaii) published in Graphics
 * Gems (Acadamic Press 1990).
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#include <os.h>
#include <libndls.h>
#include <stdint.h>


#define WAIT(n, name) \
  asm volatile ( \
    "ldr r0, ="n"\n" \
    "_loop_"name":\n" \
    "subs r0, r0, #1\n" \
    "bne _loop_"name"\n" \
    : : : "r0" );


//! Generator for 17-bit numbers.
#define LFSR_XOR_MASK (0x12000)

//! Holds the LFSR value
uint32_t seq;


/**
 * \brief dissolving routine
 *
 * This effect uses a sequence generator which generates all possible
 * 15-bit numbers in a pseudo-randomized and very fast way. This method is
 * known as Linear Feedback Shift Register (LFSR). It's also used in a
 * number of electronical devices. The mathematics behind LFSRs are
 * Galois fields (finite fields). 
 *
 * The 3 LSB are used as bitmask (well, somehow, given how bit instructions
 * work) and the 12 MSB are used as byte offset of a pixel within the LCD
 * memory.
 *
 * The sequence generator is initialized with number 1. If number 1 is
 * generated again by the sequence generator all 15-bit numbers have been
 * generated exactly once !
 *
 * Normally dissolving is used to clear a screen or to fade-in a picture.
 * For demonstration purposes my implementation just inverts the picture.
 *
 * The XOR value (0x6000) generates all 15-bit numbers. Here is a list of
 * XOR values for other sequences, up to 32-bit numbers.
 *
 * BITWIDTH  MASK
 *
 *    2      0x03
 *    3      0x06
 *    4      0x0C
 *    5      0x14
 *    6      0x30
 *    7      0x60
 *    8      0xB8
 *    9      0x0110
 *   10      0x0240
 *   11      0x0500
 *   12      0x0CA0
 *   13      0x1B00
 *   14      0x3500
 *   15      0x6000      <--- used in this demo
 *   16      0xB400
 *   17      0x00012000
 *   18      0x00020400
 *   19      0x00072000
 *   20      0x00090000
 *   21      0x00140000
 *   22      0x00300000
 *   23      0x00400000
 *   24      0x00D80000
 *   25      0x01200000
 *   26      0x03880000
 *   27      0x07200000
 *   28      0x09000000
 *   29      0x14000000
 *   30      0x32800000
 *   31      0x48000000
 *   32      0xA3000000
 *
 * In general the code for such a pseudo-random sequence generator will look
 * like the following (SOMETYPE being an unsigned variable type long enough to
 * hold all generated values):
 *
 * SOMETYPE seq = 1;
 * do {
 *     if (seq & 1) seq = (seq>>1) ^ XOR_MASK;
 *     else         seq = seq>>1;
 * }
 * while (seq != 1);
 *
 *
 *
 * NOTE: an optimization hint is to define "static" internal routines which
 *       are used just once or twice.
 *       This enables the optimizer to completely inline them, it if can
 *       determine that it will result in better code.
 *
 */
static int dissolve(void) {

    do {
        // This code assumes 4 bits per pixel.
        uint32_t byteoffset = seq >> 1;

        if (byteoffset < SCREEN_BYTES_SIZE) {
            uint8_t *p = SCREEN_BASE_ADDRESS + byteoffset;
            uint8_t val = *p;

            if (seq & 1) {
                val = (val & 0xF0) | (15 - (val & 0x0F));
            }
            else {
                val = ((15 << 4) - (val & 0xF0)) | (val & 0x0F);
            }

            *p = val;
        }

        if (seq & 1) seq = (seq >> 1) ^ LFSR_XOR_MASK;
        else         seq = seq >> 1;

        WAIT("0x17F", "loop_slowdown");

        if (isKeyPressed(KEY_NSPIRE_ESC)) {
            // THIS RETURNS !
            return 1;
        }
    }
    while (seq != 1);

    return 0;
}


//! main routine: where all the fun starts...
int main(void) {
    int ret = 0;

    seq = 1234567891;

    show_msgbox(
        "dissolve - Dissolve effect demo",
        "------------------------\n"
        "  Dissolve effect demo\n"
        "    © 2000-2010 TICT\n"
        "------------------------\n"
        "ESC   - Exit"
    );

    //-----------------------------------------------------------
    // wait a little bit and do the dissolve/XOR effect treatment
    //-----------------------------------------------------------
    while (ret == 0) {
        WAIT("0xFFFFF", "wait");

        ret = dissolve();
    }

    return 0;
}
